"""
the example illustrates how to perform check for oofun derivatives
"""
from openopt import NLP,  oofun,  oovar

v0 = oovar('v0', [-4.0, -15.0]) # 2nd arg (if provided) is start value
v1 = oovar('v2', (3, 4), lb=[-2, 0.0], ub=[-1.5, 1]) # lb, ub are upper / lower bounds (optional)
# also, Python tuple or numpy ndarray or matrix can be used instead of Python list

f0 = oofun(lambda y, z: [y[0]**2 + 4 * y[1]**2 + z[0]**2, y[0]**2 + 2 * y[1]**2 + z[1]**2], input = [v0, v1])
f0.d = lambda y, z: [[2*y[0], 8*y[1], 2*z[0], 0], [2*y[0], 4*y[1], 0, 15+2*z[1]]]
# incorrect derivative 15+2*z[1], should be 2*z[1]

# objective function
F = oofun(lambda t: 2*t.sum()+1, input = f0)

p = NLP(F)

# let's check f0 derivatives
p.checkOOFunDerivatives(f0) 
# also, you could check some oofun instances:
# p.checkOOFunDerivatives(f0, f1,f2 ,,, fk) 

"""
output: 
f num   i,j: df[i]/dx[j]   user-supplied     numerical               RD
    7              1 / 3         +2.300e+01     +8.000e+00              3
max(abs(f_user - f_numerical)) = 15.0000012329
(is registered in f number 7)
"""

# optional code for the example: solve the problem
f0.d = lambda y, z: [[2*y[0], 8*y[1], 2*z[0], 0], [2*y[0], 4*y[1], 0, 2*z[1]]]
r = p.solve('ralg')







