from numpy import hstack, ceil, floor, log10, inf, tile, argmax, abs, asfarray, ravel, any
from string import rjust, ljust
from ooFun import oofun

class autocreate:
    def __init__(self):pass
    
separator= 64 * '*'

def checkGradient(p, fun_, *args, **kwargs):
    """
    fun_ is string name like 'df', 'dc', 'dh'
    for to avoid problems with scaling, the func must be called before p.solve() has been executed.
    lately I intend to fix the problem. // D.
    """
    if not isinstance(fun_, oofun): # classic style
        p.__prepare__()
        if not getattr(p.userProvided, fun_): 
            p.warn("you haven't analitical gradient provided for " + fun_ + ', turning derivatives check for it off...')
            return

        if len(args)>0:
            if len(args)>1 or 'x' in kwargs.keys():
                p.err('checkd<func> funcs can have single argument x only (then x should be absent in kwargs )')
            xCheck = asfarray(args[0])
        elif 'x' in kwargs.keys():
            xCheck = asfarray(kwargs['x'])
        else:
            xCheck = asfarray(p.x0)

        if 'maxViolation' in kwargs.keys():
            setattr(p, 'maxViolation', kwargs['maxViolation'])

        if p.isObjFunValueASingleNumber and fun_ == 'df': 
            useDoubleColumn = False
        else:
            useDoubleColumn = True
        

        setattr(p.userProvided, fun_, False)
        info_numerical = getattr(p, fun_)(xCheck, ind=None, ignorePrev=True)
        setattr(p.userProvided, fun_, True)

        info_user = getattr(p, fun_)(xCheck, ignorePrev=True)
    else: # named variables style
        if not hasattr(fun_, 'd') or fun_.d is None:
            p.warn("you haven't analitical gradient provided for oofun " + fun_.name + ', turning derivatives check for it off...')
            return
        info_user = fun_.D(p.x0)
        delattr(fun_, 'd_key_prev')
        tmp_d, tmp_fixed, fun_.d, fun_.fixed = fun_.d, fun_.fixed, None, False
        info_numerical = fun_.D(p.x0)
        fun_.d, fun_.fixed = tmp_d, tmp_fixed
        if fun_.outputTotalLength>1: 
            useDoubleColumn = True
        else:
            useDoubleColumn = False

    if info_numerical.shape != info_user.shape:
        p.err('user-supplied gradient ' + fun_ + ' has other size than the one, obtained numerically: '+ \
        str(info_numerical.shape) + ' expected, ' + str(info_user.shape) + ' obtained')
        return inf

    Diff = 1 - (info_user+1e-8)/(info_numerical + 1e-8) # 1e-8 to suppress zeros
    log10_RD = log10(abs(Diff)/p.maxViolation+1e-150)

    d = hstack((info_user.reshape(-1,1), info_numerical.reshape(-1,1), Diff.reshape(-1,1)))
    
    if isinstance(fun_, oofun): s = 'for oofun ' + fun_.name
    else: s = fun_
    print('OpenOpt checks user-supplied gradient ' + s + ' of shape ' + str(info_user.shape))
    print('according to:')
    if isinstance(fun_, oofun):
        print('    oofun.diffInt = ' + str(fun_.diffInt)) # TODO: ADD other parameters: allowed epsilon, maxDiffLines etc
        print('    |1 - info_user/info_numerical| < oofun.maxViolation = '+ str(fun_.maxViolation))
    else:
        print('    prob.diffInt = ' + str(p.diffInt)) # TODO: ADD other parameters: allowed epsilon, maxDiffLines etc
        print('    |1 - info_user/info_numerical| < prob.maxViolation = '+ str(p.maxViolation))

    if any(abs(Diff) >= p.maxViolation):
        ss = '    '
        if isinstance(fun_, oofun): S = 'f'
        else: S = fun_[1:]
            
        if useDoubleColumn:
            ss = ' i,j: d' + S + '[i]/dx[j]'

        s = S + ' num  ' + ss + '   user-supplied     numerical               RD'
        print(s)

    ns = ceil(log10(d.shape[0]))
    counter = 0
    fl_info_user = info_user.flatten()
    fl_info_numerical = info_numerical.flatten()
    if len(Diff.shape) == 1:
        Diff = Diff.reshape(-1,1)
        log10_RD = log10_RD.reshape(-1,1)
    for i in xrange(Diff.shape[0]):
        for j in xrange(Diff.shape[1]):
            if abs(Diff[i,j]) < p.maxViolation: continue
            counter += 1
            k = Diff.shape[1]*i+j
            nSpaces = ns - floor(log10(k+1))+2
            if useDoubleColumn:  ss = str(i) + ' / ' + str(j)
            else: ss = ''

            if len(Diff.shape) == 1 or Diff.shape[1] == 1: n2 = 0
            else: n2 = 15
            s = '    ' + ljust('%d' % k,5) + rjust(ss, n2) + rjust('%+0.3e' % fl_info_user[k],19) + rjust('%+0.3e' % fl_info_numerical[k], 15) + rjust('%d' % int(ceil(log10_RD[i,j])), 15)
            print(s)

    diff_d = abs(d[:,0]-d[:,1])
    ind_max = argmax(diff_d)
    val_max = diff_d[ind_max]
    if any(abs(Diff) >= p.maxViolation):
        print('max(abs('  + S + '_user - ' + S + '_numerical)) = ' + str(val_max))
        print('(is registered in '+ S+ ' number ' + str(ind_max) + ')')
    else:
        print('derivatives are equal')
    #print('sum(abs('  + fun_ + '_user - ' + fun_ + '_numerical)) = ' + str(p.norm(d[:,2],1)))

    print(separator)

    #delattr(p.user, fun_)
    #delattr(p.user, fun_[1:])
    #p.x0 = asfarray(x0)
    if isinstance(fun_, oofun): 
        assert hasattr(fun_, 'same')
        assert hasattr(fun_, 'evals')
        fun_.same = 0
        fun_.evals = 0
    else:
        p.nEvals[fun_[1:]] = 0
        p.nEvals[fun_] = 0
    
    #return counter




